package com.cs427.swordfish.banking;

import osl.manager.annotations.message;
import osl.manager.Actor;
import osl.manager.ActorName;

/*
 * Account is the Actor for every account.  
 * This was made into it's own class to increase scalability.
 * It processes transactions passed to it by a Bank. 
 */
public abstract class Account extends Actor {

	// Serializable ID
	private static final long serialVersionUID = -4429477148227683686L;

	// Cents in account
	protected int cents = 0;

	// PIN for account
	private String pin;

	// Whether account is locked or not
	public boolean isLocked = false;

	// Amount of incorrect attempts with PIN
	private int counter = 0;

	// Amount of attempts with incorrect PIN before locking account
	public static final int MAX_ATTEMPTS = 3;

	// Helper transaction for transfer
	private Transfer transferTrans = null;

	private void send(Transaction trans, Response resp) {
		send(trans.getActor(), trans.getCallback(), resp);
	}

	/*
	 * Constructor Initializes account without starting balance specified (to 0
	 * cents)
	 * 
	 * @param pin Pin for this account
	 */
	public Account(String pin) {
		this(0, pin);
	}

	/*
	 * Constructor Initializes account with starting balance as specified
	 * 
	 * @param startingBalance The starting balance of the account
	 * 
	 * @param pin Pin for this account
	 */
	public Account(Integer startingBalance, String pin) {
		cents = startingBalance;
		this.pin = pin;
	}

	/*
	 * Withdraws amount specified in transaction from account
	 * 
	 * @param trans The transaction with the specified amount to withdraw
	 */
	@message
	public void withdraw(Transaction trans) {
		withdraw(trans, true);
	}

	/*
	 * Withdraws amount specified in transaction from account
	 * 
	 * @param trans The transaction with the specified amount to withdraw
	 * 
	 * @param multiples Whether to require multiples of 2000 on withdrawal
	 */
	@message
	public void withdraw(Transaction trans, Boolean multiples) {
		if (isLocked) {
			send(trans, new FailureResponse("Account is locked."));
			return;
		}
		if (!checkPin(trans.getPin())) {
			send(trans, new FailureResponse("Invalid PIN."));
			return;
		}
		if ((!multiples.booleanValue()) && (trans.getAmount() % 2000 != 0)) {
			send(trans, new FailureResponse(
					"Withdrawals must be in multiples of 20"));
			return;
		}
		if ((trans.getAmount() < 0) || (trans.getAmount() > cents)) {
			send(trans, new FailureResponse("Withdraw failed."));
		} else {
			cents -= trans.getAmount();
			send(trans, new SuccessResponse("Withdraw succeeded."));
		}
	}

	/*
	 * Deposits amount specified in transaction from account
	 * 
	 * @param trans The transaction with the specified amount to deposit
	 */
	@message
	public boolean deposit(Transaction trans) {
		if (isLocked) {
			send(trans, new FailureResponse("Account is locked."));
			return false;
		}
		if (!checkPin(trans.getPin())) {
			send(trans, new FailureResponse("Invalid PIN."));
			return false;
		}
		double amount = trans.getAmount();
		if (amount < 0) {
			send(trans, new FailureResponse("Deposit failed."));
			return false;
		} else {
			cents += amount;
			send(trans, new SuccessResponse("Deposit succeeded."));
			return true;
		}
	}

	/*
	 * Inquires balance from account
	 * 
	 * @param trans The transaction with the specified pin to validate
	 */
	@message
	public void balanceInquiry(Transaction trans) {
		if (isLocked) {
			send(trans, new FailureResponse("Account is locked."));
			return;
		}
		if (this.transferTrans != null) {
			send(trans, new FailureResponse(
					"Transaction cannot be completed at this time."));
			return;
		}
		if (!checkPin(trans.getPin())) {
			send(trans, new FailureResponse("Invalid PIN."));
			return;
		}
		send(trans, new BalanceResponse(cents));
	}

	/*
	 * Checks to see if pin is valid
	 * 
	 * @param p pin we're checking as input
	 * 
	 * @return true if the pin is valid, false if not or account locked
	 */
	public boolean checkPin(String p) {
		if (!isLocked && p.trim().equals(pin)) {
			counter = 0;
			return true;
		} else {
			counter++;
			if (counter >= MAX_ATTEMPTS)
				isLocked = true;
			return false;
		}
	}

	/*
	 * Unlocks account
	 * 
	 * @param trans Transaction with specified callback function
	 */
	@message
	public void unlock(Transaction trans) {
		counter = 0;
		isLocked = false;
		send(trans, new SuccessResponse("Account successfully unlocked!"));
	}

	/*
	 * Locks account
	 * 
	 * @param trans Transaction with specified callback function
	 */
	@message
	public void lock(Transaction trans) {
		isLocked = true;
		send(trans, new SuccessResponse("Account successfully locked!"));
	}

	/*
	 * Determines whether the account is close-able
	 * 
	 * @return whether the account is close-able
	 */
	@message
	public boolean close() {
		return new Boolean(cents == 0 && this.transferTrans == null);
	}

	/*
	 * Initiate a transfer between an account.  Asynchronous, calls transComplete as a callback.
	 * @param trans Transfer transaction object
	 * @param from Account to transfer funds FROM.
	 */
	@message
	public void transfer(Transfer trans, ActorName from) {
		if (isLocked) {
			send(trans, new FailureResponse("Destination account is locked."));
			return;
		}
		if (this.transferTrans != null) {
			send(trans, new FailureResponse(
					"Transaction cannot be completed at this time."));
			return;
		}
		try {
			this.transferTrans = trans;
			Transaction wTrans = new Transaction(self(), "transComplete",
					new Integer(0), trans.getAmount(), trans.getPin(), 0, 0.0);
			send(from, "withdraw", wTrans, new Boolean(false));
		} catch (Exception e) {
			send(trans, new FailureResponse("Transfer failed.  Sorry."));
		}
	}

	/*
	 * transComplete is the callback from the transfer handler.
	 * @param resp Response object representing the result of the transfer (success or failure)
	 */
	@message
	public void transComplete(Response resp) {
		if (resp instanceof SuccessResponse) {
			cents += this.transferTrans.getAmount();
			send(this.transferTrans, new SuccessResponse("Transfer succeeded!"));
		} else {
			send(this.transferTrans, new FailureResponse(
					"Withdraw during transfer failed."));
		}
		this.transferTrans = null;
	}

	/*
	 * Update account value based on interest rate
	 */
	public abstract void update();
}