package com.cs427.swordfish.banking;

import java.util.HashMap;

import osl.manager.Actor;
import osl.manager.ActorName;
import osl.manager.RemoteCodeException;
import osl.manager.annotations.message;

/*
 * Actor that simulates Bank transactions
 */
public class BankActor extends Actor {

	// Hash of account number to account objects(actors)
	private HashMap<Integer, ActorName> accounts = new HashMap<Integer, ActorName>();

	// global static counter that is incremented every time a new account is
	// created
	private static int accountIds = 0;

	// Serializable ID
	private static final long serialVersionUID = 1L;

	/* 
	 * Helper function we wrote to ease our lives
	 * @param trans Transaction with callback
	 * @param resp Response to send
	 */
	private void send(Transaction trans, Response resp) {
		send(trans.getActor(), trans.getCallback(), resp);
	}

	/*
	 * Creates account Actor, adds it to the hash
	 * 
	 * @param trans Transaction with relevant information
	 */
	@message
	public void createAccount(Transaction trans) throws RemoteCodeException {
		ActorName newAcct = null;
		int accountNumber = accountIds + 1;
		if (trans.getType() == Transaction.CHECKING) {
			newAcct = create(CheckingAccount.class, trans.getAmount(), trans
					.getPin());
		} else if (trans.getType() == Transaction.SAVINGS) {
			newAcct = create(SavingsAccount.class, trans.getAmount(), trans
					.getPin(), trans.getRate());
		} else {
			send(trans, new FailureResponse("Illegal account type."));
			return;
		}
		accounts.put(accountNumber, newAcct);
		accountIds++;
		send(trans, new OpenedResponse(accountNumber));
	}

	/*
	 * Closes account if it has zero balance(removes it from hash)
	 * 
	 * @param trans Transaction with relevant information
	 */
	@message
	public void closeAccount(Transaction trans) {
		ActorName acct = accounts.get(trans.getAccountNumber());
		if (acct == null) {
			send(trans, new FailureResponse("No such account."));
			return;
		}

		Boolean result;
		try {
			result = (Boolean) call(acct, "close");
		} catch (RemoteCodeException e) {
			send(trans, new FailureResponse(
					"Account not closed: nonzero balance."));
			return;
		}

		// call balance
		if (result.booleanValue()) {
			accounts.remove(trans.getAccountNumber());
			send(trans, new SuccessResponse("Account closed."));
		} else {
			send(trans, new FailureResponse(
					"Account not closed: nonzero balance."));
		}
	}

	/*
	 * Delegates withdraw to account
	 * 
	 * @param trans Transaction that contains relevant information
	 */
	@message
	public void withdraw(Transaction trans) throws RemoteCodeException {
		ActorName acct = accounts.get(trans.getAccountNumber());
		if (acct == null) {
			send(trans, new FailureResponse("No such account."));
			return;
		}

		send(acct, "withdraw", trans);
	}

	/*
	 * Delegates deposit to account
	 * 
	 * @param trans Transaction that contains relevant information
	 */
	@message
	public void deposit(Transaction trans) throws RemoteCodeException {
		ActorName acct = accounts.get(trans.getAccountNumber());
		if (acct == null) {
			send(trans, new FailureResponse("No such account."));
			return;
		}

		send(acct, "deposit", trans);
	}

	/*
	 * Delegates transfer to destination account actor
	 * 
	 * @param trans Transaction that contains relevant information
	 */
	@message
	public void transfer(Transfer trans) throws RemoteCodeException {
		ActorName from = accounts.get(trans.getFrom());
		if (from == null) {
			send(trans, new FailureResponse("No such account:"
					+ trans.getFrom()));
			return;
		}

		ActorName to = accounts.get(trans.getAccountNumber());
		if (to == null) {
			send(trans, new FailureResponse("No such account:"
					+ trans.getAccountNumber()));
			return;
		}
		send(to, "transfer", trans, from);

	}

	/*
	 * Delegates balance inquiry to account
	 * 
	 * @param trans Transaction with relevant information
	 */
	@message
    public void balanceInquiry(Transaction trans) {
		ActorName acct = accounts.get(trans.getAccountNumber());

		if (acct == null) {
			send(trans, new FailureResponse("No such account:"
					+ trans.getAccountNumber()));
		} else {
			send(acct, "balanceInquiry", trans);
		}
	}

	/*
	 * Unlocks an account by delegating to the account
	 * 
	 * @param trans Transaction with relevant information
	 */
	@message
	public void unlockAccount(Transaction trans) {
		ActorName acct = accounts.get(trans.getAccountNumber());

		if (acct == null) {
			send(trans, new FailureResponse("No such account:"
					+ trans.getAccountNumber()));
		} else {
			send(acct, "unlock", trans);
		}
	}

	/*
	 * Locks an account by delegating to the account
	 * 
	 * @param trans Transaction with relevant information
	 */
	@message
	public void lockAccount(Transaction trans) {
		ActorName acct = accounts.get(trans.getAccountNumber());

		if (acct == null) {
			send(trans, new FailureResponse("No such account:"
					+ trans.getAccountNumber()));
		} else {
			send(acct, "lock", trans);
		}
	}

}
