package com.cs427.swordfish.banking;

import java.io.Serializable;

import osl.manager.ActorName;

/*
 * Encapsulates all transactions so they can be passed from actor to actor
 */
public class Transaction implements Serializable {

	// Serializable ID
	private static final long serialVersionUID = 1L;

	// Actor to call back to
	private ActorName actor;

	// Name of function to call back to
	private String callback;

	// Account number to run transaction on
	private Integer accountNumber;

	// Amount relevant for transaction type
	private int amount;

	// PIN of account
	private String pin;

	// Type of account (Checking, Savings, NA) - as specified below
	private int type;

	// Rate of account
	private double rate;

	// Types of account
	public static final int CHECKING = 0;
	public static final int SAVINGS = 1;
	public static final int NA = 3;

	/*
	 * Constructor
	 * 
	 * @param actor Actor to call back to
	 * 
	 * @param callback Method to call back to
	 * 
	 * @param accountNumber account to run transaction on
	 * 
	 * @param amount Amount relevant to transaction
	 * 
	 * @param pin PIN for this account
	 * 
	 * @param type Type of this account
	 * 
	 * @param rate Rate of this account
	 */
	public Transaction(ActorName actor, String callback, Integer accountNumber,
			Integer amount, String pin, int type, double rate) {
		this.actor = actor;
		this.callback = callback;
		this.accountNumber = accountNumber;
		this.amount = amount;
		this.pin = pin;
		if (type == NA) {
			this.type = NA;
			this.rate = new Double(0.0);
			return;
		}
		if ((type != CHECKING) && (type != SAVINGS)) {
			throw new IllegalArgumentException("Invalid account type.");
		}
		this.type = type;
		this.rate = rate;
	}

	/*
	 * Getter for Actor
	 * 
	 * @return Actor to call back to
	 */
	public ActorName getActor() {
		return actor;
	}

	/*
	 * Getter for callback
	 * 
	 * @return Callback function
	 */
	public String getCallback() {
		return callback;
	}

	/*
	 * Getter for account number
	 * 
	 * @return account number
	 */
	public Integer getAccountNumber() {
		return accountNumber;
	}

	/*
	 * Getter for amount
	 * 
	 * @return amount
	 */
	public int getAmount() {
		return amount;
	}

	/*
	 * Getter for pin
	 * 
	 * @return pin
	 */
	public String getPin() {
		return pin;
	}

	/*
	 * Getter for type
	 * 
	 * @return type
	 */
	public int getType() {
		return type;
	}

	/*
	 * Getter for rate
	 * 
	 * @return rate
	 */
	public double getRate() {
		return rate;
	}

}

/*
 * Derived object for transfer transactions
 */
class Transfer extends Transaction {

	// Serializable ID
	private static final long serialVersionUID = 1L;

	// Account number for source
	private Integer from;

	/*
	 * Constructor
	 * 
	 * @param actor Actor to call back to
	 * 
	 * @param callback Method to call back to
	 * 
	 * @param from account that is source
	 * 
	 * @param to account that is destination
	 * 
	 * @param amount Amount relevant to transaction
	 * 
	 * @param pin PIN for this account
	 */
	public Transfer(ActorName actor, String callback, Integer from, Integer to,
			int amount, String pin) {
		super(actor, callback, to, amount, pin, NA, 0);
		this.from = from;
	}

	/*
	 * Getter for source account
	 * 
	 * @return source account number
	 */
	public Integer getFrom() {
		return from;
	}
}
