package edu.illinois.cs.cs427.clownfish.actorBanking;

import osl.manager.Actor;
import osl.manager.ActorName;
import osl.manager.RemoteCodeException;
import osl.manager.annotations.message;

public class ActorATM extends Actor{
	private ActorName bank;
	
	/**
	 * Creates an ATM for an owning bank
	 * @param b the bank that owns this ATM
	 * @throws RemoteCodeException
	 */
	public ActorATM(ActorName b) throws RemoteCodeException{
		bank = b;
	}

	/**
	 * Executes a banking transaction and returns the result of the transaction.
	 * 
	 * @param primaryAccountNum the account to perform the transaction on
	 * @param secondaryAccountNum the account number to transfer funds to (when performing a transfer)
	 * @param amount the amount to deposit/withdraw/transfer
	 * @param type the type of transaction (deposit/withdraw/transfer/inquiry/printlog/setpin)
	 * @param submittedPin the entered pin for the transaction
	 * @param newPin the new pin if transaction is setpin
	 * @return the result of the transaction; SuccessResponse, FailureResponse, or DenialResponse
	 * @throws RemoteCodeException
	 */
	@message
	private BankingResponse transaction(Integer primaryAccountNum, Integer secondaryAccountNum, Integer amount, ActorBank.transactionType type, Integer submittedPin, Integer newPin) throws RemoteCodeException {
		// get TransactionManager from bank
		ActorName transactionManager = (ActorName)call(bank, "getTransactionManager");
		
		// attempt transaction
		return (BankingResponse)call(transactionManager, "doTransaction", primaryAccountNum, secondaryAccountNum, amount, type, submittedPin, newPin);
	}
	
	/**
	 * Deposit an amount into a banking account
	 * 
	 * @param accountNum the account number of the account
	 * @param amount the amount to deposit
	 * @param submittedPin the entered pin for the transaction
	 * @return the result of the deposit; SuccessResponse, FailureResponse, or DenialResponse
	 * @throws RemoteCodeException
	 */
	@message
	public BankingResponse deposit(Integer accountNum, Integer amount, Integer submittedPin) throws RemoteCodeException  {
		return transaction(accountNum, null, amount, ActorBank.transactionType.DEPOSIT, submittedPin, null);
	}
	
	/**
	 * Withdraw an amount from a banking account
	 * 
	 * @param accountNum the account number of the account
	 * @param amount the amount to withdraw
	 * @param submittedPin the entered pin for the transaction
	 * @return the result of the withdraw; SuccessResponse, FailureResponse, or DenialResponse
	 * @throws RemoteCodeException
	 */
	@message
	public BankingResponse withdraw(Integer accountNum, Integer amount, Integer submittedPin) throws RemoteCodeException {
		return transaction(accountNum, null, amount, ActorBank.transactionType.WITHDRAW, submittedPin, null);
	}
	
	/**
	 * Get the balance of a banking account
	 * 
	 * @param accountNum the account number of the account
	 * @param submittedPin the entered pin for the transaction
	 * @return the result of the inquiry; SuccessResponse, FailureResponse, or DenialResponse
	 * @throws RemoteCodeException
	 */
	@message
	public BankingResponse balanceInquiry(Integer accountNum, Integer submittedPin) throws RemoteCodeException {
		return transaction(accountNum, null, null, ActorBank.transactionType.INQUIRY, submittedPin, null);
	}
	
	/**
	 * Transfer an amount from a primary account into a secondary account
	 * 
	 * @param primaryAccountNum the account to transfer money from
	 * @param secondaryAccountNum the account to transfer money into
	 * @param amount the amount to transfer
	 * @param submittedPin the entered pin for the transaction
	 * @return the result of the transfer; SuccessResponse, FailureResponse, or DenialResponse
	 * @throws RemoteCodeException
	 */
	@message
	public BankingResponse transfer(Integer primaryAccountNum, Integer secondaryAccountNum, Integer amount, Integer submittedPin) throws RemoteCodeException {
		return transaction(primaryAccountNum, secondaryAccountNum, amount, ActorBank.transactionType.TRANSFER, submittedPin, null);
	}
	
	/**
	 * Prints a log of the account's activity
	 * 
	 * @param primaryAccountNum the account number of the account
	 * @param submittedPin the entered pin for the transaction
	 * @return the result of the inquiry; SuccessResponse, FailureResponse, or DenialResponse
	 * @throws RemoteCodeException
	 */
	@message
	public BankingResponse printLog(Integer primaryAccountNum, Integer submittedPin) throws RemoteCodeException {
		return transaction(primaryAccountNum, null, null, ActorBank.transactionType.PRINTLOG, submittedPin, null);
	}
	
	/**
	 * Sets the pin of the account
	 * 
	 * @param primaryAccountNum the account number of the account
	 * @param newPin the entered pin for the transaction
	 * @param submittedPin the entered pin for the transaction
	 * @return the result of the inquiry; SuccessResponse, FailureResponse, or DenialResponse
	 * @throws RemoteCodeException
	 */
	@message
	public BankingResponse setPin(Integer primaryAccountNum, Integer newPin, Integer submittedpin) throws RemoteCodeException {
		return transaction(primaryAccountNum, null, null, ActorBank.transactionType.SETPIN, submittedpin, newPin);
	}
	
	/**
	 * Open a bank account
	 * 
	 * @param initAmount the initial amount to put in the bank account
	 * @param submittedPin the entered pin for the transaction
	 * @return the account number of the account that was created
	 * @throws RemoteCodeException
	 */
	@message
	public Integer openAccount(Integer initAmount, Integer submittedPin) throws RemoteCodeException {
		return (Integer) call(bank, "openAccount", initAmount, submittedPin);
	}
}
