package edu.illinois.cs.cs427.clownfish.actorBanking;
import java.util.Date;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.LinkedList;

import osl.manager.Actor;
import osl.manager.ActorName;
import osl.manager.RemoteCodeException;
import osl.manager.annotations.message;

public class ActorAccount extends Actor{

	private Integer accountNumber;
	private Integer balance;
	private LinkedList <LogEntry> log;
	private Integer pin;
	
	/**
	 * Populates the new ActorAccount with an account number and initial balance and initial pin
	 * 
	 * @param accountNumber the account number for the new account
	 * @param startingBalance the initial balance for the new account
	 * @param pin the initial pin for the new account
	 */
	public ActorAccount(Integer accountNumber, Integer startingBalance, Integer pin) {
		this.accountNumber = accountNumber;
		this.balance = startingBalance;
		this.pin = pin;
		this.log = new LinkedList<LogEntry>();
		log.add(new LogEntry(getDateTime(), Integer.toString(balance), 
				"Opened", "None"));
	}
	
	/**
	 * Get the current time and date in a simple string format
	 * 
	 * @return string the current date in yyyy/MM/dd HH:mm:ss format
	 */
	private String getDateTime() {
		DateFormat dateFormat = new SimpleDateFormat("yyyy/MM/dd HH:mm:ss");
		Date date = new Date();
		return dateFormat.format(date);
	}
	 
	/**
	 * Checks a submitted pin against the account's pin
	 * 
	 * @param submittedPin the pin to be checked
	 * @return boolean true if submitted is equal to actual pin; false otherwise
	 */
	@message
	public Boolean checkPin(Integer submittedPin) {
		return (submittedPin.equals(this.pin));
	}
	
	/**
	 * Sets a new pin for the account
	 * 
	 * @param newPin the new pin for the account
	 * @return boolean true always
	 */
	@message
	public Boolean setPin(Integer newPin) {
		this.pin = newPin;
		return true;
	}
	
	/**
	 * Gets the balance on the account
	 * 
	 * @return Integer the balance
	 */
	@message
	public Integer getBalance() {
		return balance;
	}
	
	/**
	 * Gets the pin of the account
	 * 
	 * @return Integer the pin
	 */
	@message
	public Integer getPin() {
		return pin;
	}
	
	/**
	 * Deposit an amount into the account
	 * 
	 * @param amount the amount to deposit
	 * @param fromTrans flag if the call is from transfer or not
	 * @return Boolean true if successful; false otherwise
	 */
	@message
	public Boolean deposit(Integer amount, Boolean fromTrans) {
		//If your balance won't overflow
		if(balance < (Integer.MAX_VALUE - amount)) {
			balance += amount;
			if (!fromTrans){
				log.add(new LogEntry(getDateTime(), Integer.toString(balance), 
						"Deposit", "None"));
			}
			return true;
		}
		else {
			return false;
		}
	}
	
	/**
	 * Withdraw an amount from the account
	 * 
	 * If the account has insufficient funds to withdraw amount, no amount is withdrawn.
	 * 
	 * @param amount the amount to withdraw
	 * @param fromTrans flag if the call is from transfer or not
	 * @return Boolean true if successful; false otherwise
	 */
	@message
	public Boolean withdraw(Integer amount, Boolean fromTrans) {
		if (balance < amount){
			return false;
		}
		else {
			balance -= amount;
			if (false == fromTrans){
				log.add(new LogEntry(getDateTime(), Integer.toString(balance), 
						"Withdraw", "None"));
			}
			return true;
		}
	}
	
	/**
	 * Transfer an amount from this account to another account
	 * 
	 * If the account has insufficient funds to transfer amount, no amount is transferred.
	 * 
	 * @param secondaryAccountName the account actor to transfer funds to
	 * @param amount the amount to transfer
	 * @return Boolean true if successful; false otherwise
	 * @throws RemoteCodeException
	 */
	@message
	public Boolean transferTo(ActorName secondaryAccountName, Integer amount) throws RemoteCodeException {
		Boolean withdrawSuccess = this.withdraw(amount, true);
		if(withdrawSuccess) {
			Boolean depositSuccess = (Boolean) call(secondaryAccountName, "deposit", amount, true);
			if(!depositSuccess) {
				//Roll Back Transaction
				this.deposit(amount, true);
				return false;
			}
			else {
				String date = getDateTime();
				call(secondaryAccountName, "addToLog", date, Integer.toString(amount),"Transfer From", Integer.toString(this.accountNumber));
				Integer otherAccount = (Integer) call(secondaryAccountName, "getAccountNumber");
				log.add(new LogEntry(getDateTime(), Integer.toString(balance), 
							"Transfer To", Integer.toString(otherAccount)));
				return true;
			}
		}
		else {
			return false;
		}
	}
	
	/**
	 * Gets the account number of the account
	 * 
	 * @return Integer the account number
	 */
	@message
	public Integer getAccountNumber() {
		return accountNumber;
	}
	
	/**
	 * Logs a transaction to the account's log
	 * 
	 * @param date the time and date
	 * @param amt the amount of the transaction
	 * @param type the transaction type
	 * @param acct the other account involved if type is transfer
	 * @return Integer the account number
	 */
	@message
	public void addToLog(String date, String amt, String type, String acct) {
		log.add(new LogEntry(date, amt, type, acct));
	}
	
	/**
	 * Prints the log of the account to the ATM's screen in table format
	 *
	 * @return boolean always returns true
	 */
	@message
	public boolean printLog() {
		System.out.println("Log for account number " + accountNumber);
		System.out.println();
		System.out.println("Date\t\t\tAmount   Transaction Type  Second Account");
		System.out.println("--------------------------------------------------------------------------");
		Integer count = 0;
		for(LogEntry cur:log) {
			System.out.print(cur.getDate() + " \t ");
			System.out.print(cur.getAmount() + " \t   ");
			System.out.print(cur.getType() + " \t          ");
			System.out.println(cur.getAcct());
			count++;
		}
		printSummary();
		return true;
	}
	
	/**
	 * Prints a summary of the Account's log including totals of deposits/withdraws
	 * 
	 */
	@message
	private void printSummary() {
		Integer deposits = 0;
		Integer depositAmount = 0;
		Integer withdrawals = 0;
		Integer withdrawalAmount = 0;
		for(LogEntry cur:log) {
			if(cur.getType().equals("Deposit") || cur.getType().equals("Transfer From")) {
				deposits++; 
				depositAmount += Integer.parseInt(cur.getAmount());
			}
			if(cur.getType().equals("Withdraw") || cur.getType().equals("Transfer To")) {
				withdrawals++; 
				withdrawalAmount += Integer.parseInt(cur.getAmount());
			}
		}
		System.out.println("Number of deposits: " + deposits);
		System.out.println("Total amount deposited: " + depositAmount);
		System.out.println("Number of withdrawals: " + withdrawals);
		System.out.println("Total amount withdrawn: " + withdrawalAmount);

		
	}
}
