package edu.illinois.cs.cs427.clownfish.actorBanking;
import java.util.HashMap;
import java.util.Vector;

import osl.manager.Actor;
import osl.manager.ActorName;
import osl.manager.RemoteCodeException;
import osl.manager.annotations.message;

public class ActorBank extends Actor{
	
	public enum transactionType {
		DEPOSIT, 
		WITHDRAW,
		INQUIRY,
		TRANSFER,
		PRINTLOG, 
		SETPIN
	}
	
	private final int INITIAL_TRANSACTION_MANAGER_POOL_SIZE = 5;
	
	private int nextAccountNum = 0;
	private int curTransactionManager = 0;
	private Integer transactionFee; 
	private HashMap<Integer, ActorName> accounts = new HashMap<Integer, ActorName>();
	private Vector<ActorName> transactionManagers = new Vector<ActorName>();
	
	/**
	 * Creates a bank with initial transaction fee of 0
	 */
	public ActorBank() {
		this.transactionFee = 0;
	}
	
	/**
	 * Sets a new transaction fee for the bank
	 * 
	 * @param transactionFee the initial transaction fee for the new bank
	 */
	public ActorBank(Integer transactionFee) {
		this.transactionFee = transactionFee;
	}
	
	/**
	 * Sets up the transaction managers for the bank
	 * 
	 * @throws RemoteCodeException
	 * @throws SecurityCodeException
	 */
	@message
	public void initialize() throws SecurityException, RemoteCodeException {
		for (int i = 0; i < INITIAL_TRANSACTION_MANAGER_POOL_SIZE; i++)
		{
			ActorName managerName = create(TransactionManager.class, self());
			transactionManagers.add(managerName);
		}
	}
	
	/**
	 * Opens and creates a new account
	 * Should only be used with a "call", not a "send". 
	 * 
	 * @param initAmount the initial amount for the account
	 * @param pin the initial pin for the account
	 * @return Account number of new account
	 * @throws SecurityException
	 * @throws RemoteCodeException
	 */
	@message
	public Integer openAccount(Integer initAmount, Integer pin) throws SecurityException, RemoteCodeException {
		Integer accountNum = nextAccountNum;
		nextAccountNum++;
		ActorName curName = create(ActorAccount.class, accountNum, initAmount, pin);
		accounts.put(accountNum, curName);
		return accountNum;
	}
	
	/**
	 * Checks a submitted account number against the bank's collection of accounts
	 * 
	 * @param accountNum the account number to be checked
	 * @return boolean true
	 */
	private boolean accountExists(Integer accountNum) {
		return accounts.containsKey(accountNum);
	}
	
	/**
	 * Gets the next transaction manager
	 * 
	 * @return ActorName of transaction manager
	 */
	private ActorName getNextTransactionManager() {
		ActorName transactionManager = transactionManagers.get(curTransactionManager);
		curTransactionManager = (curTransactionManager + 1) % transactionManagers.size();
		
		return transactionManager;
	}
	
	/**
	 * Wrapper for getNextTransactionManager
	 * 
	 * @return ActorName of transaction manager
	 */
	@message
	public ActorName getTransactionManager() {
		return getNextTransactionManager();
	}
	
	/**
	 * Returns the actor account of the account number submitted
	 * 
	 * @param accountNum the account number to be returned
	 * @return ActorName of the requested account
	 */
	@message
	public ActorName getAccountActorName(Integer accountNum) {
		if(!accountExists(accountNum)) {
			throw new RuntimeException("Account " + accountNum + " Does Not Exist");
		}
		else {
			return accounts.get(accountNum);
		}
	}
	
	/**
	 * Sets a new transaction fee for the bank
	 * 
	 * @param transactionFee the new transaction fee for the bank
	 * @return boolean true always
	 */
	@message
	public void modifyTransactionFee(Integer newFee) {
		this.transactionFee = newFee;
	}
	
	/**
	 * Gets the transaction fee for the bank
	 * 
	 * @return Integer the transaction fee for the bank
	 */
	@message
	public Integer getFee() {
		return transactionFee;
	}
}
