package edu.illinois.cs.cs427.clownfish.actorBanking;

import java.util.HashMap;

import osl.manager.Actor;
import osl.manager.ActorName;
import osl.manager.RemoteCodeException;
import osl.manager.annotations.message;

public class TransactionManager extends Actor {
	
	private ActorName owningBank;
	private HashMap<Integer, ActorName> accountMap;
	
	public TransactionManager(ActorName owningBank)
	{
		this.owningBank = owningBank;
		accountMap = new HashMap<Integer, ActorName>();
	}
	
	/**
	 * 
	 * @param accountNum
	 * @return ActorName for the requested account; null if account does not exist.
	 */
	@message
	private ActorName getAccountActorName(Integer accountNum) throws RemoteCodeException {
		ActorName accountName = accountMap.get(accountNum);
		
		if (null == accountName)
		{
			//System.out.println("Account not found in transaction manager cache");
			accountName = (ActorName) call(owningBank, "getAccountActorName", accountNum);
		}
		
		return accountName;
	}

	/**
	 * Executes a banking transaction and returns the result of the transaction.
	 * 
	 * @param primaryAccountNum the account to perform the transaction on
	 * @param secondaryAccountNum the account number to transfer funds to (when performing a transfer)
	 * @param amount the amount to deposit/withdraw/transfer
	 * @param type the type of transaction (deposit/withdraw/transfer/inquiry)
	 * @return a BankingResponse indicating the result of the transaction
	 * @throws RemoteCodeException
	 */
	@message
	public BankingResponse doTransaction(Integer primaryAccountNum, Integer secondaryAccountNum, Integer amount, ActorBank.transactionType type, Integer submittedPin, Integer newPin) throws RemoteCodeException {
		Integer fee = (Integer) call(owningBank, "getFee");
		
		ActorName requestedAccountName = getAccountActorName(primaryAccountNum);
		if (null == requestedAccountName)
		{
			return new FailureResponse();
		}

		Boolean success = false;
		Integer returnedAmount = -1;
		BankingResponse response;
		
		Boolean authPin = (Boolean) call(requestedAccountName, "checkPin", submittedPin);
		if (!authPin){
			response = new DenialResponse();
			return response;
		}
		
		// Perform the transaction on the account
		if (type == ActorBank.transactionType.SETPIN){
			success = (Boolean) call(requestedAccountName, "setPin", newPin);
		}
		else if(type == ActorBank.transactionType.DEPOSIT) {
			success = (Boolean) call(requestedAccountName, "deposit", amount - fee, false);
		}
		else if(type == ActorBank.transactionType.WITHDRAW) {
			success = (Boolean) call(requestedAccountName, "withdraw", amount + fee, false);
		}
		else if(type == ActorBank.transactionType.INQUIRY) {
			returnedAmount = (Integer) call(requestedAccountName, "getBalance");
			if(returnedAmount >= 0) {
				success = true;
			}
			else {
				success = false;
			}
		}
		else if(type == ActorBank.transactionType.TRANSFER) {
			ActorName secondaryAccountName = (ActorName) call(owningBank, "getAccountActorName", secondaryAccountNum);
			success = (Boolean) call(requestedAccountName, "transferTo", secondaryAccountName, amount);
		}
		else if(type == ActorBank.transactionType.PRINTLOG) {
			success = (Boolean) call(requestedAccountName, "printLog");
		}
		
		else {
			throw new RuntimeException("Unrecognized transaction type!");
		}
		
		// Package the response based on the results of the transaction
		if(success) {
			if(type == ActorBank.transactionType.INQUIRY) {
				response = new BalanceResponse(returnedAmount);
			}
			else {
				response = new SuccessResponse();
			}
		}
		else {
			response = new FailureResponse();
		}

		return response;
	}
}
