/*
 * Authors:
 * 
 * Jeffrey Dennhardt
 * Kurt Hendle
 * Yuzuko Nakamura
 * Jurand Nogiec
 * 
 * Written Fall 2009
 */

package edu.illinois.cs427.Banking;

import java.util.HashMap;
import java.util.Random;

import osl.manager.annotations.message;
import osl.manager.Actor;
import osl.manager.ActorName;
import osl.manager.RemoteCodeException;

public class Bank extends Actor {
	private static final long serialVersionUID = 5825848367673594014L;

	private HashMap<String, Account> accounts = new HashMap<String, Account>();
	private int bankNumber;
	private int numBanks;

	/**
	 * this default constructor is needed by ActorFoundry to function correctly
	 */
	public Bank() { }
	
	/**
	 * constructor used in the scalable design
	 * @param bankNum
	 * @param numBanks
	 */
	public Bank(int bankNum, int numBanks){
		this.bankNumber = bankNum;
		this.numBanks = numBanks;
	}
	
	public int getBankNum() {
		return this.bankNumber;
	}
	
	/**
	 * Make an account number starting with A randomly assigned.
	 * 
	 * @return
	 */
	private String randomAcctNr() {
		// generate random account number
		Random r = new Random();
		int randint = r.nextInt(100000);
		return "A" + ((Integer) randint).toString();
	}
	
	/**
	 * Generates account number within a given bank's range.
	 * 
	 * @param bankNum
	 * @return the account number
	 */
	private String randomAcctNr(int bankNum) {
		Random r = new Random();
		int randint = r.nextInt(100000/numBanks) + bankNum * (100000/numBanks);
		return Integer.toString(randint);
	}

	/**
	 * add an account with a unique number to the accounts set.
	 * 
	 * @return
	 * @throws RemoteCodeException
	 */
	@message
	public Response addAccount() throws RemoteCodeException {
		String newAcctNr;
		// check if unique
		do {
			newAcctNr = randomAcctNr();
		} while (accounts.get(newAcctNr) != null);

		accounts.put(newAcctNr, new Account(newAcctNr));

		return (new Response(true, newAcctNr));
	}
	
	/**
	 * adds an account within the appropriate range of account numbers this bank manages
	 * @return
	 */
	@message
	public Response scalableAddAccount() {
		String  newAcctNr;
		// check if unique
		do {
			newAcctNr = randomAcctNr(bankNumber);
		} while (accounts.get(newAcctNr) != null);

		accounts.put(newAcctNr, new Account(newAcctNr));

		return (new Response(true, newAcctNr));
	}

	/**
	 * close the account given the name
	 * 
	 * @param acctName
	 * @return whether or not it successfully closed.
	 * @throws RemoteCodeException
	 */
	@message
	public Response closeAccount(String acctName) throws RemoteCodeException {
		if (accounts.containsKey(acctName)) {
			Account acct = accounts.get(acctName);
			if(acct.getBalance()==0) {
				accounts.remove(acctName);
				return (new Response(true));
			}
		}
		return (new Response(false, acctName));
	}
	/**
	 * takes care of actual deposit at the bank (calls to print just to verify
	 * that it works visually)
	 * 
	 * @param acctName
	 * @param amount
	 * @throws RemoteCodeException
	 */
	@message
	public Response deposit(String acctName, Double amount)
			throws RemoteCodeException {
		if (accounts.containsKey(acctName)) {
			Account acct = accounts.get(acctName);
			acct.deposit(amount);
			return (new Response(true, acct.getBalance()));
		} else {
			return (new Response(false, 0.0));
		}
	}

	/**
	 * takes care of actual withdraw at the bank (calls to print just to verify
	 * that it works visually)
	 * 
	 * @param acctName
	 * @param amount
	 * @throws RemoteCodeException
	 */
	@message
	public Response withdraw(String acctName, Double amount)
			throws RemoteCodeException {
		if (accounts.containsKey(acctName)) {
			Account acct = accounts.get(acctName);
			acct.withdraw(amount);
			return (new Response(true, acct.getBalance()));

		} else {
			return (new Response(false, 0.0));

		}
	}

	/**
	 * takes care of actual transfer at the bank (calls to print just to verify
	 * that it works visually)
	 * 
	 * @param acctNameFrom
	 *            Account to transfer money FROM
	 * @param acctNameTo
	 *            Account to transfer money TO
	 * @param amount
	 * @throws RemoteCodeException
	 */
	@message
	public Response transfer(String acctNameFrom, String acctNameTo, Double amount)
			throws RemoteCodeException {
		if (accounts.containsKey(acctNameFrom)
				&& accounts.containsKey(acctNameTo)) {
			Account acctFrom = accounts.get(acctNameFrom);
			if (acctFrom.getBalance() >= amount) {
				acctFrom.withdraw(amount);
				Account acctTo = accounts.get(acctNameTo);
				acctTo.deposit(amount);
				return (new Response(true));
			}
		}
		return (new Response(false));
	}
	
	/**
	 * Handles a transfer if 1 or multiple banks are involved. Withdraw from the source account
	 * and deposit into the destination account at its bank.
	 * @param acctNameFrom - account from which money being transferred
	 * @param acctNameTo - account where money is being deposited
	 * @param toBank - destination account's bank
	 * @param amount - amount to transfer
	 */
	@message
	public Response scalableTransfer(String acctNameFrom, String acctNameTo, ActorName toBank, 
			Double amount) throws RemoteCodeException {
		if(accounts.containsKey(acctNameFrom)) {
			Account acctFrom = accounts.get(acctNameFrom);
			if(acctFrom.getBalance() >= amount) {
				acctFrom.withdraw(amount);
				Response didWork = (Response) call(toBank, "deposit", acctNameTo, amount);
				return didWork;
			}
		}
		return (new Response(false));
	}

	/**
	 * balance inquiry at the bank, returns the account balance
	 * 
	 * @param acctName
	 * @return
	 * @throws RemoteCodeException
	 */
	@message
	public Response balanceInquiry(String acctName) throws RemoteCodeException {
		if (accounts.containsKey(acctName)) {
			return (new Response(true, accounts.get(acctName).getBalance()));
		} else {
			return (new Response(false, 0.0));
		}
	}
}
