package rainbow_fish.project;

import java.util.HashMap;
import java.util.Map;

import osl.manager.Actor;
import osl.manager.ActorName;
import osl.manager.RemoteCodeException;
import osl.manager.annotations.message;

/**
 * ATMActor represents an ATM object that a customer uses to perform a
 * transaction on an account owned by the customer.
 * 
 */
public class ATMActor extends Actor {

	private static final long serialVersionUID = 1L;
	private ActorName bank;

	/**
	 * Constructor for this <tt>ATMActor</tt>. The associated <tt>BankActor</tt>
	 * is specified by the bank_in parameter.
	 * 
	 * @param bank_in
	 *            An <tt>ActorName</tt> representing the <tt>BankActor</tt>
	 *            associated with <tt>this</tt>
	 */
	public ATMActor(ActorName bank_in) {
		this.bank = bank_in;
	}

	/**
	 * Performs a deposit to the account specified by primaryAccount, with the
	 * amount specified by amount, represented as an <tt>Amount</tt> object.
	 * 
	 * @param primaryAccount
	 *            A <tt>String</tt> representing the account id to deposit into.
	 * @param amount
	 *            An <tt>Amount</tt> representing the amount to deposit.
	 * @return A <tt>Response</tt> object representing the success/failure of
	 *         the deposit.
	 * @throws RemoteCodeException
	 */
	@message
	public Response deposit(String primaryAccount, Amount amount)
			throws RemoteCodeException {
		Response transaction = (Response) call(bank, "getDepositTransaction",
				primaryAccount);
		if (transaction instanceof FailureResponse)
			return transaction;
		else
			return (Response) call(((TransactionResponse) transaction)
					.getAccount(), "deposit", amount);
	}

	/**
	 * Performs a withdraw from the account specified by primaryAccount, with
	 * the amount specified by amount, represented as an <tt>Amount</tt> object.
	 * 
	 * @param primaryAccount
	 *            A <tt>String</tt> representing the account id to withdraw
	 *            from.
	 * @param amount
	 *            An <tt>Amount</tt> representing the amount to withdraw.
	 * @return A <tt>Response</tt> object representing the success/failure of
	 *         the deposit.
	 * @throws RemoteCodeException
	 */
	@message
	public Response withdraw(String primaryAccount, Amount amount)
			throws RemoteCodeException {
		Response transaction = (Response) call(bank, "getWithdrawTransaction",
				primaryAccount);
		if (transaction instanceof FailureResponse)
			return transaction;
		else
			return (Response) call(((TransactionResponse) transaction)
					.getAccount(), "withdraw", amount);
	}

	/**
	 * Performs a transfer from primaryAccount to secondaryAccount, with the
	 * amount specified by amount, represented as an <tt>Amount</tt> object.
	 * 
	 * @param primaryAccount
	 *            A <tt>String</tt> representing the account id to withdraw
	 *            from.
	 * @param secondaryAccount
	 *            A <tt>String</tt> representing the account id to deposit into.
	 * @param amount
	 *            An <tt>Amount</tt> representing the amount to transfer.
	 * @return A <tt>Response</tt> object representing the success/failure of
	 *         the deposit.
	 * @throws RemoteCodeException
	 */
	@message
	public Response transfer(String primaryAccount, String secondaryAccount,
			Amount amount) throws RemoteCodeException {
		Response transaction = (Response) call(bank, "getTransferTransaction",
				primaryAccount, secondaryAccount);
		if (transaction instanceof FailureResponse)
			return transaction;
		else
			return (Response) call(((TransactionResponse) transaction)
					.getAccount(), "transfer", amount);
	}

	/**
	 * Gets the balance of the account specified by primaryAccount.
	 * 
	 * @param primaryAccount
	 *            A <tt>String</tt> representing the account id to retrieve
	 *            balance.
	 * @return A <tt>BalanceResponse</tt> object encapsulating the amount that
	 *         is in the account.
	 * @throws RemoteCodeException
	 */
	@message
	public Response getBalance(String primaryAccount)
			throws RemoteCodeException {
		Response transaction = (Response) call(bank, "getBalanceTransaction",
				primaryAccount);
		if (transaction instanceof FailureResponse)
			return transaction;
		else
			return (Response) call(((TransactionResponse) transaction)
					.getAccount(), "getBalance");
	}
}
