package rainbow_fish.project;

import java.io.Serializable;

/**
 * Amount object represents a balance amount of an AccountActor in the Bank
 * system. Amount is represented in cents.
 */
public class Amount implements Comparable<Amount>, Serializable {

	private static final long serialVersionUID = 1L;

	public static Amount ZERO_AMOUNT = new Amount(0);

	private long cents;

	/**
	 * Returns a new <tt>Amount object</tt> with value represented by this
	 * Amount object.
	 * 
	 * @return <tt>this</tt>
	 */
	public Amount getValue() {
		return new Amount(this.cents);
	}

	/**
	 * Creates a new <tt>Amount</tt> object with the given number of cents.
	 * 
	 * @param cents
	 *            the number of cents in the amount.
	 *            Trying to create a negative amount will create a 0 amount.
	 */
	public Amount(long cents) {
		if (cents <=0)
			this.cents = 0;
		else
			this.cents = cents;
	}

	/**
	 * Creates a new <tt>Amount</tt> object with 0 value.
	 */
	public Amount() {
		this(0);
	}

	/**
	 * Creates a new <tt>Amount</tt> object with the given number of dollars.
	 * 
	 * @param dollars
	 *            the number of dollars in the amount.
	 */
	public Amount(double dollars) {
		this((long) (dollars * 100));
	}

	/**
	 * Returns an <tt>Amount</tt> whose value is the sum of <tt>this</tt> and
	 * <tt>other</tt>.
	 * 
	 * @param other
	 *            the amount to add.
	 * @return <tt>this + other</tt>
	 */
	public Amount add(Amount other) {
		return new Amount(cents + other.cents);
	}

	/**
	 * Returns an <tt>Amount</tt> whose value is <tt>this</tt> minus and
	 * <tt>other</tt>.
	 * 
	 * @param other
	 *            the amount to subtract.
	 * @return <tt>this - other</tt>
	 */
	public Amount subtract(Amount other) {
		return new Amount(cents - other.cents);
	}

	/**
	 * Compares values of two <tt>Amount</tt> objects.
	 * 
	 * @param other
	 *            the <tt>Amount</tt> to be compared.
	 * @return the value <tt>0</tt>, negative value, or a positive value, if
	 *         this <tt>Amount</tt> is equal to the other, less than the other,
	 *         or greater than the other, respectively.
	 */
	public int compareTo(Amount other) {
		if (cents < other.cents)
			return -1;
		else if (cents > other.cents)
			return 1;
		else
			return 0;
	}

	/**
	 * Compares this <tt>Amount</tt> to the given object.
	 * 
	 * @param o
	 *            the object to compare with.
	 * @return <tt>true</tt> if the objects are the same; <tt>false</tt>
	 *         otherwise.
	 */
	public boolean equals(Object o) {
		if (o != null && o.getClass() == this.getClass()) {
			return cents == ((Amount) o).cents;
		}
		return false;
	}

	/**
	 * Returns the hash code for this <tt>Amount</tt>.
	 * 
	 * @return the hash code.
	 */
	public int hashCode() {
		return (int) (cents ^ (cents >>> 32));
	}

	/**
	 * Returns the string representation for this <tt>Amount</tt>.
	 * 
	 * @return the value formatted as "DOLLARS.CC".
	 */
	public String toString() {
		return (cents / 100) + "." + (cents % 100);
	}
}
