package rainbow_fish.project;

import java.util.HashMap;
import java.util.Map;

import osl.manager.Actor;
import osl.manager.ActorName;
import osl.manager.RemoteCodeException;
import osl.manager.annotations.message;

/**
 * BankActor represents a Bank institution where accounts are maintained at a
 * central office. Each BankActor serves as an intermediary from where ATM
 * request a transaction on an account, and the bank returns a transaction to be
 * performed if the account is valid.
 */
public class BankActor extends Actor {

	private static final long serialVersionUID = 1L;

	private int nextAccount = 0;

	private Map<String, ActorName> accounts = new HashMap<String, ActorName>();

	/**
	 * Checks to see if the account with the id specified by accountId exists or
	 * not.
	 * 
	 * @param accountId
	 *            A <tt>String</tt> representing the account id of the account
	 *            to be checked.
	 * @return A <tt>Response</tt> object representing the result.
	 *         <tt>SuccessResponse</tt> if account exists,
	 *         <tt>FailureResponse</tt> otherwise.
	 */
	@message
	public Response accountExists(String accountId) {
		if (accounts.get(accountId) != null)
			return new SuccessResponse();
		else
			return new FailureResponse("Account doesn't exist\n");
	}

	/**
	 * Opens a new account with a zero balance
	 * 
	 * @return id of the created account object.
	 */
	@message
	public String openAccount() throws RemoteCodeException {
		String accountId = Integer.toString(++nextAccount);
		ActorName someAccount = create(AccountActor.class);
		accounts.put(accountId, someAccount);
		return accountId;
	}

	/**
	 * Closes an account, removing it from the bank. Operation fails if account
	 * does not exist or has a non-zero balance
	 * 
	 * @param accountId
	 * @return <tt>true</tt> if the account successfully closed.
	 */
	@message
	public Boolean closeAccount(String accountId) throws RemoteCodeException {
		if (accounts.containsKey(accountId)
				&& (call(accounts.get(accountId), "getBalance")
						.equals(Amount.ZERO_AMOUNT))) {
			accounts.remove(accountId);
			return new Boolean(true);
		}
		return new Boolean(false);
	}

	/**
	 * Returns the <tt>Account</tt> object for the given account id.
	 * 
	 * @param accountId
	 *            account id
	 * @return the <tt>Account</tt> object for the given account id.
	 *         <tt>null</tt> if the bank does not have this account.
	 */
	@message
	private ActorName getAccount(String accountId) throws RemoteCodeException {
		return accounts.get(accountId);
	}

	/**
	 * Creates a <tt>TransactionActor</tt> for the accountId specified to
	 * perform a deposit, for the account with accountId specified, then return
	 * a <tt>TransactionResponse</tt> encapsulating the <tt>ActorName</tt> for
	 * the <tt>TransactionActor</tt> created.
	 * 
	 * @param accountId
	 *            A <tt>String</tt> representing the account id to perform the
	 *            deposit
	 * @return A <tt>TransactionResponse</tt> encapsulating the
	 *         <tt>ActorName</tt> for the <tt>TransactionActor</tt> that will
	 *         perform the deposit
	 * @throws RemoteCodeException
	 */
	@message
	public Response getDepositTransaction(String accountId)
			throws RemoteCodeException {
		ActorName account = getAccount(accountId);
		if (account != null) {
			return new TransactionResponse(create(TransactionActor.class,
					account));
		} else {
			return new FailureResponse("Account " + accountId
					+ " does not exist.\n");
		}
	}

	/**
	 * Creates a <tt>TransactionActor</tt> for the accountId specified to
	 * perform a withdraw, for the account with accountId specified, then return
	 * a <tt>TransactionResponse</tt> encapsulating the <tt>ActorName</tt> for
	 * the <tt>TransactionActor</tt> created.
	 * 
	 * @param accountId
	 *            A <tt>String</tt> representing the account id to perform the
	 *            withdrawal
	 * @return A <tt>TransactionResponse</tt> encapsulating the
	 *         <tt>ActorName</tt> for the <tt>TransactionActor</tt> that will
	 *         perform the withdrawal
	 * @throws RemoteCodeException
	 */
	@message
	public Response getWithdrawTransaction(String accountId)
			throws RemoteCodeException {
		ActorName account = getAccount(accountId);
		if (account != null) {
			return new TransactionResponse(create(TransactionActor.class,
					account));
		} else {
			return new FailureResponse("Account " + accountId
					+ " does not exist.\n");
		}
	}

	/**
	 * Creates a <tt>TransactionActor</tt> to perform a transfer, from
	 * primaryAccountId to secondaryAccountId, then return a
	 * <tt>TransactionResponse</tt> encapsulating the <tt>ActorName</tt> for the
	 * <tt>TransactionActor</tt> created.
	 * 
	 * @param primaryAccountId
	 *            A <tt>String</tt> representing the accountId of the account to
	 *            withdraw from.
	 * @param secondaryAccountId
	 *            A <tt>String</tt> representing the accountId of the account to
	 *            deposit into.
	 * @return A <tt>TransactionResponse</tt> encapsulating the
	 *         <tt>ActorName</tt> for the <tt>TransactionActor</tt> that will
	 *         perform the transfer
	 * @throws RemoteCodeException
	 */
	@message
	public Response getTransferTransaction(String primaryAccountId,
			String secondaryAccountId) throws RemoteCodeException {
		ActorName primaryAccount = getAccount(primaryAccountId);
		ActorName secondaryAccount = getAccount(secondaryAccountId);
		if (primaryAccount != null && secondaryAccount != null) {
			return new TransactionResponse(create(TransactionActor.class,
					primaryAccount, secondaryAccount));
		} else
			return new FailureResponse("Invalid account name(s) supplied.\n");
	}

	/**
	 * Creates a <tt>TransactionActor</tt> for the accountId specified to return
	 * balance from, and return a <tt>TransactionResponse</tt> encapsulating the
	 * <tt>ActorName</tt> of the <tt>TransactionActor</tt> created.
	 * 
	 * @param accountId
	 *            A <tt>String</tt> representing the accountId to check balance
	 *            from.
	 * @return A <tt>TransactionResponse</tt> encapsulating the
	 *         <tt>ActorName</tt> for the <tt>TransactionActor</tt> that will
	 *         perform the get balance.
	 * @throws RemoteCodeException
	 */
	@message
	public Response getBalanceTransaction(String accountId)
			throws RemoteCodeException {
		ActorName account = getAccount(accountId);
		if (account != null) {
			return new TransactionResponse(create(TransactionActor.class,
					account));
		} else
			return new FailureResponse("Account " + accountId
					+ " does not exist.\n");
	}
}
