package rainbow_fish.project;

import osl.manager.ActorName;

/**
 * Response is a class to abstract responses/return values to every call/sends
 * to Actors. It can be used to indicate a successful/failed message processing,
 * or to encapsulate ActorNames or Amount depending on the called Actor and type
 * of message.
 */
public abstract class Response {
}

class FailureResponse extends Response {
	private String failureMessage;

	/**
	 * Creates a <tt>FailureResponse</tt> object, with the specified
	 * failureMessage.
	 * 
	 * @param failureMessage
	 *            A <tt>String</tt> representing the failureMessage for this
	 *            <tt>FailureResponse</tt>
	 */
	public FailureResponse(String failureMessage) {
		this.failureMessage = failureMessage;
	}

	/**
	 * Gets the failure message for this <tt>FailureResponse</tt>
	 * 
	 * @return A <tt>String</tt> specifying the failureMessage.
	 */
	public String getMessage() {
		return failureMessage;
	}
}

class SuccessResponse extends Response {
}

class BalanceResponse extends Response {
	private Amount balance;

	/**
	 * Creates a <tt>BalanceResponse</tt> with the specified balance, which is
	 * represented by an <tt>Amount</tt> object.
	 * 
	 * @param balance
	 *            An <tt>Amount</tt> object representing the balance to be
	 *            contained in this <tt>BalanceResponse</tt>
	 */
	public BalanceResponse(Amount balance) {
		this.balance = balance;
	}

	/**
	 * Gets the balance represented by this <tt>BalanceResponse</tt>.
	 * 
	 * @return An <tt>Amount</tt> object representing the balance specified in
	 *         this <tt>BalanceResponse</tt>
	 */
	public Amount getBalance() {
		return balance;
	}
}

class TransactionResponse extends SuccessResponse {
	private ActorName account;

	/**
	 * Creates a <tt>TransactionResponse</tt> object to be used by the ATM to
	 * perform a transaction.
	 * 
	 * @param account
	 *            An <tt>ActorName</tt> for the <tt>TransactionActor</tt> that
	 *            will perform the transaction.
	 */
	public TransactionResponse(ActorName account) {
		this.account = account;
	}

	/**
	 * Gets the <tt>ActorName</tt> for the <tt>TransactionActor</tt> to be used
	 * by the ATM to perform a transaction.
	 * 
	 * @return An <tt>ActorName</tt> for the <tt>TransactionActor</tt> to
	 *         perform a transaction.
	 */
	public ActorName getAccount() {
		return this.account;
	}
}
