package rainbow_fish.project;

import osl.manager.Actor;
import osl.manager.ActorName;
import osl.manager.RemoteCodeException;
import osl.manager.annotations.message;

/**
 * TransactionActor represents a teller/agent that handles a particular
 * transaction. It allows for scalability of transaction calls to the bank, and
 * guarantees that the transaction is performed correctly, and allows for atomic
 * transaction calls for each ATMActor. After the transaction, it will send a
 * response if it is successful or not.
 */
public class TransactionActor extends Actor {

	private static final long serialVersionUID = 1030595902702952346L;

	private ActorName primaryAccount;
	private ActorName secondaryAccount;

	/**
	 * Creates a <tt>TransactionActor</tt> to perform a transfer, from
	 * <tt>AccountActor</tt> specified by <tt>
	 * ActorName</tt> pri to sec.
	 * 
	 * @param pri
	 *            An <tt>ActorName</tt> for the <tt>AccountActor</tt> to
	 *            withdraw from.
	 * @param sec
	 *            An <tt>ActorName</tt> for the <tt>AccountActor</tt> to deposit
	 *            into.
	 */
	public TransactionActor(ActorName pri, ActorName sec) {
		this.primaryAccount = pri;
		this.secondaryAccount = sec;
	}

	/**
	 * Creates a <tt>TransactionActor</tt> to perform a withdraw from/ deposit
	 * into, the <tt>AccountActor</tt> specified by <tt>ActorName</tt> pri.
	 * 
	 * @param pri
	 *            An <tt> ActorName</tt> for the <tt>AccountActor</tt> to
	 *            withdraw from/deposit into.
	 */
	public TransactionActor(ActorName pri) {
		this.primaryAccount = pri;
	}

	/**
	 * Performs a deposit into primaryAccount, with the amount specified.
	 * 
	 * @param amount
	 *            An <tt>Amount</tt> object representing the amount to deposit.
	 * @return A <tt>Response</tt> object specifying if the deposit was
	 *         successful. Since deposits will always be successfull, it always
	 *         returns a <tt>SuccessResponse</tt>.
	 * @throws RemoteCodeException
	 */
	@message
	public Response deposit(Amount amount) throws RemoteCodeException {
		call(primaryAccount, "deposit", amount);
		return new SuccessResponse();
	}

	/**
	 * Performs a withdrawal from primaryAccount, with the amount specified.
	 * 
	 * @param amount
	 *            An <tt>Amount</tt> object representing the amount to withdraw.
	 * @return A <tt>Response</tt> object specifying if the withdraw was
	 *         successful. <tt>SuccessResponse</tt> if it is,
	 *         <tt>FailureResponse</tt> otherwise.
	 * @throws RemoteCodeException
	 */
	@message
	public Response withdraw(Amount amount) throws RemoteCodeException {
		Integer result = (Integer) call(primaryAccount, "withdraw", amount);
		if (result.intValue() == -1)
			return new FailureResponse("Balance is less than amount.\n");
		else
			return new SuccessResponse();
	}

	/**
	 * Performs a transfer from primaryAccount to secondaryAccount.
	 * 
	 * @param amount
	 *            An <tt>Amount</tt> object specifying the amount to transfer.
	 * @return A <tt>Response</tt> object indicating the status of the transfer.
	 *         <tt>SuccessResponse</tt> if so, <tt>
	 * FailureResponse</tt> otherwise.
	 * @throws RemoteCodeException
	 */
	@message
	public Response transfer(Amount amount) throws RemoteCodeException {
		Integer result = (Integer) call(primaryAccount, "withdraw", amount);
		if (result.intValue() == -1)
			return new FailureResponse(
					"Balance in primary account is less than amount.\n");
		else {
			call(secondaryAccount, "deposit", amount);
			Amount balanceOne = (Amount)call(primaryAccount,"getBalance");
			Amount balanceTwo = (Amount)call(secondaryAccount,"getBalance");
			System.out.println(balanceOne.toString());
			System.out.println(balanceTwo.toString());
			return new SuccessResponse();
		}
	}

	/**
	 * Returns the balance of primaryAccount, encapsulated within a
	 * <tt>BalanceResponse</tt> object.
	 * 
	 * @return A <tt>BalanceResponse</tt> object encapsulating the amount of
	 *         balance within primaryAccount.
	 * @throws RemoteCodeException
	 */
	@message
	public Response getBalance() throws RemoteCodeException {
		Amount amount = (Amount) call(primaryAccount, "getBalance");
		BalanceResponse response = new BalanceResponse(amount);
		return response;
	}
}
